# noinspection PyUnresolvedReferences
# - Convenience pass-through from "private" sub-module.
from smqtk_detection._defaults import DFLT_DETECTION_FACTORY  # noqa: F401
from .object_detector import ObjectDetector, ImageMatrixObjectDetector  # noqa: F401



import abc
import hashlib

import numpy

from typing import Hashable, Set, Iterator, Dict, Tuple, Any, Type, TypeVar

from smqtk_core import Plugfigurable
from smqtk_image_io.interfaces.image_reader import ImageReader
from smqtk_dataprovider import ContentTypeValidator
from smqtk_core.configuration import (
    make_default_config,
    from_config_dict,
    to_config_dict
)

from smqtk_classifier._defaults import DFLT_CLASSIFIER_FACTORY
from smqtk_detection._defaults import DFLT_DETECTION_FACTORY

from smqtk_detection.utils.bbox import AxisAlignedBoundingBox
from smqtk_detection.interfaces.detection_element import DetectionElement
from smqtk_detection.detection_element_factory \
    import DetectionElementFactory
from smqtk_dataprovider.interfaces.data_element import DataElement
from smqtk_classifier.classification_element_factory import ClassificationElementFactory

ImMatObDet = TypeVar("ImMatObDet", bound="ImageMatrixObjectDetector")


class ObjectDetector (Plugfigurable, ContentTypeValidator, metaclass=abc.ABCMeta):
    """
    Abstract interface to an object detection algorithm.

    An object detection algorithm is one that can take in data and output zero
    or more detection elements, where each detection represents a spatial
    region in the data.

    This high level interface only requires detection element returns (spatial
    bounding-boxes with associated classification elements).
    """

    __slots__ = ()

    @staticmethod
    def _gen_detection_uuid(data_uuid: str, bbox: AxisAlignedBoundingBox, labels: Any) -> Hashable:
        """
        Local standard for producing the UUID of a DetectionElement based on
        parent data, component bounding box and classification labels.

        :param str data_uuid:
            UUID of parent data element (checksum hash string) this detection
            is derived from.
        :param smqtk.representation.AxisAlignedBoundingBox bbox:
            Detection bounding box instance.
        :param labels:
            Sequence of string classification labels.

        :return: Detection UUID string that is the SHA1 checksum of the
            component data.

        """
        # noinspection PyStringFormat
        # - uses variadic expansion which fools current linter.
        hashable = data_uuid + \
            '{}{}{}{}'.format(*(bbox.min_vertex.tolist() +
                                bbox.max_vertex.tolist())) + \
            ''.join(sorted(map(str, labels)))
        return hashlib.sha1(bytes(hashable, "utf8")).hexdigest()

    def detect_objects(
            self,
            data_element: DataElement,
            de_factory: DetectionElementFactory = DFLT_DETECTION_FACTORY,
            ce_factory: ClassificationElementFactory = DFLT_CLASSIFIER_FACTORY
    ) -> Iterator[DetectionElement]:
        """
        Detect objects in the given data.

        UUIDs of detections are based on the hash produced from the combination
        of:

        - Detection bounding-box bounding coordinates

        - Classification label set predicted for a bounding box.

        :param smqtk.representation.DataElement data_element:
            Source data from which to detect objects within.
        :param smqtk.representation.DetectionElementFactory de_factory:
            Factory for generating DetectionElement instances. The default
            factory yields MemoryClassificationElement instances.
        :param smqtk.representation.ClassificationElementFactory ce_factory:
            Factory for generating ClassificationElement instances for
            detections. The default factory yields MemoryClassificationElement
            instances.

        :raises ValueError: Given data element content was not of a valid
            content type that this class reports as valid for object detection.

        :return: Iterator over result DetectionElement instances as generated
            by the given DetectionElementFactory, containing classification
            elements as generated by the given ClassificationElementFactory.
        :rtype: collections.abc.Iterable[smqtk.representation.DetectionElement]

        """
        self.raise_valid_element(data_element)

        # We know that the UUID of a DataElement should be a checksum of sorts,
        # so we can generally assume a string-cast is unique preserving.
        de_uuid = str(data_element.uuid())

        type_str = 'object detection classification'
        for bbox, c_map in self._detect_objects(data_element):
            # Determine UUID of detection from bbox and classification labels
            det_uuid = self._gen_detection_uuid(de_uuid, bbox, c_map.keys())

            ce = ce_factory.new_classification(type_str, det_uuid)
            ce.set_classification(c_map)

            de = de_factory.new_detection(det_uuid).set_detection(bbox, ce)
            yield de

    @abc.abstractmethod
    def _detect_objects(self, data: DataElement) -> Iterator[Tuple[AxisAlignedBoundingBox, Dict[Hashable, float]]]:
        """
        Internal method that defines the generation of paired bounding boxes
        and classification maps for detected objects in the given data.

        :param smqtk.representation.DataElement data:
            Source data (DataElement) from which to detect objects within.

        :return: Iterable over paired ``AxisAlignedBoundingBox`` and
            classification map for detected objects. The returned
            "classification map" should follow the format described by
            ``smqtk.representation.ClassificationElement``: dictionary where
            keys are classification labels and values are classification
            probabilities.
        :rtype:
            collections.abc.Iterator[(smqtk.representation.AxisAlignedBoundingBox,
                                      dict[collections.abc.Hashable, float])]

        """


class ImageMatrixObjectDetector (ObjectDetector, metaclass=abc.ABCMeta):
    """
    Class of object detectors that operate over the pixel matrix of an image.

    This sub abstract class standardizes the use of an
    :class:`smqtk.algorithms.ImageReader` algorithm to read an image file's
    pixels as well as determine which image formats are valid input elements.
    There is a special exception of :class:`.MatrixDataElement` types as they
    directly provide a matrix.

    We define an alternate abstract method for implementing classes to define:
    ``_detect_objects_matrix``.  This method is given a numpy ndarray instance
    for the implementing class to utilize.  The return requirements are the same
    as the ``_detect_objects`` method.
    """

    @classmethod
    def get_default_config(cls) -> dict:
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as arguments,
        turning those argument names into configuration dictionary keys. If any
        of those arguments have defaults, we will add those values into the
        configuration dictionary appropriately. The dictionary returned should
        only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this class.

        :return: Default configuration dictionary for the class.
        :rtype: dict
        """
        default = super(ImageMatrixObjectDetector, cls).get_default_config()
        default['image_reader'] = make_default_config(ImageReader.get_impls())
        return default

    @classmethod
    def from_config(  # type: ignore
            cls: Type[ImMatObDet],
            config_dict: dict,
            merge_default: bool = True
    ) -> ImMatObDet:
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless an instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: ImageMatrixObjectDetector
        """
        # Shallow copy
        config_dict = dict(config_dict)

        config_dict['image_reader'] = from_config_dict(
            config_dict.get('image_reader', {}), ImageReader.get_impls()
        )

        return super(ImageMatrixObjectDetector, cls).from_config(
            config_dict, merge_default=merge_default
        )

    def __init__(self, image_reader: ImageReader) -> None:
        """
        An image matrix object detector must have a method of converting a
        DataElement into an image pixel matrix so this interface broadly
        requires an :class:`smqtk.algorithms.ImageReader` instance.

        :param smqtk.algorithms.ImageReader image_reader:
            ImageReader algorithm instance for reading image matrices from
            DataElements.
        """
        super(ImageMatrixObjectDetector, self).__init__()
        self._image_reader = image_reader

    @abc.abstractmethod
    def get_config(self) -> dict:
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        In the most cases, this involves naming the keys of the dictionary
        based on the initialization argument names as if it were to be passed
        to the constructor via dictionary expansion.  In some cases, where it
        doesn't make sense to store some object constructor parameters are
        expected to be supplied at as configuration values (i.e. must be
        supplied at runtime), this method's returned dictionary may leave those
        parameters out. In such cases, the object's ``from_config``
        class-method would also take additional positional arguments to fill in
        for the parameters that this returned configuration lacks.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        return {
            'image_reader': to_config_dict(self._image_reader),
        }

    def valid_content_types(self) -> Set[str]:
        """
        :return: A set valid MIME types that are "valid" within the implementing
            class' context.
        :rtype: set[str]
        """
        return self._image_reader.valid_content_types()

    def is_valid_element(self, data_element: DataElement) -> bool:
        """
        Check if the given DataElement instance reports a content type that
        matches one of the MIME types reported by ``valid_content_types``.

        This override uses our stored :class:`ImageReader` algorithm instance to
        define what :class:`DataElement` instances are valid.

        :param smqtk.representation.DataElement data_element:
             Data element instance to check.

        :return: True if the given element has a valid content type as reported
            by ``valid_content_types``, and False if not.
        :rtype: bool
        """
        return self._image_reader.is_valid_element(data_element)

    def _detect_objects(self, data: DataElement) -> Iterator[Tuple[AxisAlignedBoundingBox, Dict[Hashable, float]]]:
        """
        Internal method that defines the generation of paired bounding boxes
        and classification maps for detected objects in the given data.

        This ``ImageMatrixObjectDetector`` implementation ensures that the data
        element is converted to a :class:`numpy.ndarray` before passing the result
        matrix along to the :func:`_detect_objects_matrix` method for the
        implementing class to define.

        :param smqtk.representation.DataElement data:
            Source data (DataElement) from which to detect objects within.

        :return: Iterable over paired ``AxisAlignedBoundingBox`` and
            classification map for detected objects. The returned
            "classification map" should follow the format described by
            ``smqtk.representation.ClassificationElement``: dictionary where
            keys are classification labels and values are classification
            probabilities.
        :rtype:
            collections.abc.Iterator[(smqtk.representation.AxisAlignedBoundingBox,
                                      dict[collections.abc.Hashable, float])]

        """
        return self._detect_objects_matrix(
            self._image_reader.load_as_matrix(data)
        )

    @abc.abstractmethod
    def _detect_objects_matrix(self, mat: numpy.ndarray) \
            -> Iterator[Tuple[AxisAlignedBoundingBox, Dict[Hashable, float]]]:
        """
        Internal method to be implemented that defines the generation of paired
        bounding boxes and classification maps for detected objects in the given
        image matrix data.

        :param numpy.ndarray mat:
            Image pixel matrix to detect objects within.

        :return: Iterable over paired ``AxisAlignedBoundingBox`` and
            classification map for detected objects. The returned
            "classification map" should follow the format described by
            ``smqtk.representation.ClassificationElement``: dictionary where
            keys are classification labels and values are classification
            probabilities.
        :rtype:
            collections.abc.Iterator[(smqtk.representation.AxisAlignedBoundingBox,
                                      dict[collections.abc.Hashable, float])]
        """



"""
Default values and instances for the ObjectDetection interface.
"""
# - Using the same default factory for ObjectDetector as the Classifier
#   interface.
# - Providing classifier default here for convenience.
# noinspection PyProtectedMember
from smqtk_classifier._defaults import DFLT_CLASSIFIER_FACTORY  # noqa: F401, lgtm[py/unused-import]
from smqtk_detection.detection_element_factory import DetectionElementFactory
from smqtk_detection.impls.detection_element.memory \
    import MemoryDetectionElement

DFLT_DETECTION_FACTORY = DetectionElementFactory(
    MemoryDetectionElement, {}
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smqtk_detection',
 'smqtk_detection.impls',
 'smqtk_detection.impls.detection_element',
 'smqtk_detection.interfaces',
 'smqtk_detection.utils']

package_data = \
{'': ['*']}

install_requires = \
['smqtk-classifier>=0.19.0',
 'smqtk-core>=0.18.0',
 'smqtk-dataprovider>=0.16.0',
 'smqtk-image-io>=0.16.2']

entry_points = \
{'smqtk_plugins': ['smqtk_detection.impls.detection_element.memory = '
                   'smqtk_detection.impls.detection_element.memory']}

setup_kwargs = {
    'name': 'smqtk-detection',
    'version': '0.16.0',
    'description': 'Algorithms, data structures and utilities around performing detection of inputs',
    'long_description': '## Intent\n\n## Documentation\n\nDocumentation for SMQTK is maintained at\n[ReadtheDocs](https://smqtk.readthedocs.org), including\n[build instructions](https://smqtk.readthedocs.io/en/latest/installation.html)\nand [examples](https://smqtk.readthedocs.org/en/latest/examples/overview.html).\n\nAlternatively, you can build the sphinx documentation locally for the most\nup-to-date reference (see also: [Building the Documentation](\nhttps://smqtk.readthedocs.io/en/latest/installation.html#building-the-documentation)):\n```bash\n# Navigate to the documentation root.\ncd docs\n# Install dependencies and build Sphinx docs.\npip install -r readthedocs-reqs.txt\nmake html\n# Open in your favorite browser!\nfirefox _build/html/index.html\n```\n',
    'author': 'Kitware, Inc.',
    'author_email': 'smqtk-developers@kitware.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Kitware/SMQTK-Detection',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import abc

from smqtk_core import Plugfigurable
from smqtk_detection.exceptions import NoDetectionError
from smqtk_core.dict import merge_dict
from typing import Hashable
from smqtk_detection.utils.bbox import AxisAlignedBoundingBox
from smqtk_classifier.interfaces.classification_element import ClassificationElement
from typing import Dict, Any, Tuple, Type, Optional


class DetectionElement (Plugfigurable):
    """
    Representation of a spatial detection.
    """

    __slots__ = ('_uuid',)

    @classmethod
    def get_default_config(cls) -> Dict[str, Any]:
        # Override from Configurable.
        default = super(DetectionElement, cls).get_default_config()
        # Remove runtime positional argument(s).
        del default['uuid']
        return default

    # noinspection PyMethodOverriding
    @classmethod
    def from_config(  # type: ignore
            cls: Type["DetectionElement"],
            config_dict: Dict[Any, Any],
            uuid: Hashable,
            merge_default: bool = True
            ) -> "DetectionElement":
        """
        Override of
        :meth:`smqtk.utils.configuration.Configurable.from_config` with the
        added runtime argument ``uuid``. See parent method documentation for
        details.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param collections.abc.Hashable uuid:
            UUID to assign to the produced DetectionElement.

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: DetectionElement

        """
        # Override from Configurable
        # Handle passing of runtime positional argument(s).
        if merge_default:
            config_dict = merge_dict(cls.get_default_config(), config_dict)
        config_dict['uuid'] = uuid
        return super(DetectionElement, cls).from_config(config_dict,
                                                        merge_default=False)

    def __init__(self, uuid: Hashable) -> None:
        """
        Initialize a new detection element with the given ``uuid``.

        All DetectionElement classes will take a ``uuid`` parameter as the
        first positional argument. This parameter is not configurable and is
        only specified at runtime. Implementing classes should not include
        ``uuid`` in ``get_config`` returns.

        :param collections.abc.Hashable uuid:
            Unique ID reference of the detection.

        """
        super(DetectionElement, self).__init__()
        self._uuid = uuid

    __hash__ = None  # type: ignore

    def __eq__(self, other: Any) -> bool:
        """
        Equality of two detections is defined by their equal spatial overlap
        AND their equivalent classification.

        When one element does not contain detection information but the other
        does, the two elements are of course considered NOT equal.
        If *neither* elements contain detection information, they are defined
        as NOT equal (undefined).

        :param DetectionElement other: Other detection element.
        :return: True if the two detections are equal in spacial overlap and
            classification.
        """
        try:
            s_bb, s_ce = self.get_detection()
            o_bb, o_ce = other.get_detection()
            return s_bb == o_bb and s_ce == o_ce
        except NoDetectionError:
            return False

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    def __repr__(self) -> str:
        # using "{{...}}" to skip .format activation.
        return "{:s}{{uuid: {}}}".format(self.__class__.__name__, self._uuid)

    def __nonzero__(self) -> bool:
        """
        A DetectionElement is considered non-zero if ``has_detection`` returns
        True. See method documentation for details.

        :return: True if this instance is non-zero (see above), false
            otherwise.
        :rtype: bool
        """
        return self.has_detection()

    __bool__ = __nonzero__

    @property
    def uuid(self) -> Hashable:
        return self._uuid

    #
    # Abstract methods
    #

    @abc.abstractmethod
    def __getstate__(self) -> dict:
        return {
            '_uuid': self._uuid,
        }

    @abc.abstractmethod
    def __setstate__(self, state: dict) -> None:
        self._uuid = state['_uuid']

    @abc.abstractmethod
    def has_detection(self) -> bool:
        """
        :return: Whether or not this container currently contains a valid
            detection bounding box and classification element (must be
            non-zero).
        :rtype: bool
        """

    @abc.abstractmethod
    def get_bbox(self) -> Optional[AxisAlignedBoundingBox]:
        """
        :return: The spatial bounding box of this detection.
        :rtype: smqtk.representation.AxisAlignedBoundingBox

        :raises NoDetectionError: No detection AxisAlignedBoundingBox set yet.
        """

    @abc.abstractmethod
    def get_classification(self) -> Optional[ClassificationElement]:
        """
        :return: The classification element of this detection.
        :rtype: smqtk.representation.ClassificationElement

        :raises NoDetectionError: No detection ClassificationElement set yet or
            the element is empty.
        """

    @abc.abstractmethod
    def get_detection(self) -> Tuple[AxisAlignedBoundingBox, ClassificationElement]:
        """
        :return: The paired spatial bounding box and classification element of
            this detection.
        :rtype: (smqtk.representation.AxisAlignedBoundingBox,
                 smqtk.representation.ClassificationElement)

        :raises NoDetectionError: No detection AxisAlignedBoundingBox and
            ClassificationElement set yet.

        """

    @abc.abstractmethod
    def set_detection(self, bbox: AxisAlignedBoundingBox, classification_element: ClassificationElement) \
            -> "DetectionElement":
        """
        Set a bounding box and classification element to this detection
        element.

        :param smqtk.representation.AxisAlignedBoundingBox bbox:
            Spatial bounding box instance.

        :param smqtk.representation.ClassificationElement classification_element:
            The classification of this detection.

        :raises ValueError: No, or invalid, AxisAlignedBoundingBox or
            ClassificationElement was provided.

        :returns: Self
        :rtype: DetectionElement

        """
